#include "chord_common.h"

#include <pthread.h>
#include <gmp.h>
#include <string.h>
#include <error.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/uio.h>
#include <arpa/inet.h>
#include <openssl/evp.h>

#define HASH_SIZE 20
#define NUM_SUCCESSORS 5

static int chord_join(const struct chord_opts *options,
		struct chord_network *network)
{
	const struct addrinfo hints =  {
		.ai_family = AF_INET,
		.ai_socktype = SOCK_STREAM,
		.ai_protocol = IPPROTO_TCP,
		.ai_flags = 0,
		.ai_addrlen = 0,
		.ai_addr = NULL,
		.ai_canonname = NULL,
		.ai_next =  NULL
	};
	struct addrinfo *address_list, *address;
	int peer_sfd, retval;
	unsigned char last_node;

	retval = getaddrinfo(options->node, options->service,
			&hints, &address_list);

	if (retval) {
		retval = -1;
		goto out;
	}
	else
		retval = 0;

	for (address = address_list; address; address = address->ai_next) {
		peer_sfd = socket(address->ai_family, address->ai_socktype,
				address->ai_protocol);

		if (peer_sfd == -1)
			continue;

		if (!connect(peer_sfd, address->ai_addr, address->ai_addrlen))
			break;

		if (close(peer_sfd) == -1)
			error(EXIT_FAILURE, errno, "failed to close socket");
	}
	if (!address) {
		freeaddrinfo(address_list);
		retval = -1;
		goto out;
	}

	freeaddrinfo(address_list);

	pthread_rwlock_wrlock(network->finger_table_lock);
	pthread_rwlock_rdlock(&network->local_node_lock);

	retval = chord_next_node(peer_sfd, network->local_node.hash,
			network->hash_size, network->finger_table, &last_node);

	if (close(peer_sfd) == -1)
		error(EXIT_FAILURE, errno, "could not close socket");

	while (!last_node && retval != -1) {
		peer_sfd = chord_connect(network->finger_table);
		if (peer_sfd == -1)
			break;

		retval = chord_next_node(peer_sfd, network->local_node.hash,
				network->hash_size, network->finger_table,
				&last_node);

		if (close(peer_sfd) == -1)
			error(EXIT_FAILURE, errno, "could not close socket");
	}

	chord_rwlock_unlock(&network->local_node_lock);
	chord_rwlock_unlock(network->finger_table_lock);

	out:
	return retval;
}

static int chord_init(const struct chord_opts *options,
		struct chord_network *network)
{
	const struct addrinfo hints =  {
		.ai_family = AF_INET,
		.ai_socktype = SOCK_STREAM,
		.ai_protocol = IPPROTO_TCP,
		.ai_flags = AI_PASSIVE,
		.ai_addrlen = 0,
		.ai_addr = NULL,
		.ai_canonname = NULL,
		.ai_next =  NULL
	};
	struct addrinfo *address_list, *address;
	struct sockaddr_in *local_addr;
	int retval, i;

	retval = getaddrinfo(options->local_node, options->local_service,
			&hints, &address_list);
	if (retval)
		return -1;

	for (address = address_list; address; address = address->ai_next) {
		network->ssocketfd = socket(address->ai_family,
				address->ai_socktype, address->ai_protocol);

		if (network->ssocketfd == -1)
			continue;

		if (!bind(network->ssocketfd, address->ai_addr,
				address->ai_addrlen))
			break;

		if (close(network->ssocketfd) == -1) {
			freeaddrinfo(address_list);
			error(EXIT_FAILURE, errno, "failed to close socket");
		}
	}

	if (!address) {
		freeaddrinfo(address_list);
		return -1;
	}

	local_addr = (struct sockaddr_in *) address->ai_addr;

	if (pthread_rwlock_init(&network->local_node_lock, NULL))
		return -1;

	network->local_node.address = local_addr->sin_addr.s_addr;
	network->local_node.port = local_addr->sin_port;

	freeaddrinfo(address_list);

	hash_address(network);

	for (i = 0; i < network->hash_size * 8; i++) {
		if (pthread_rwlock_init(network->finger_table_lock + i, NULL))
			return -1;

		chord_nodecpy(network->finger_table + i, &network->local_node,
				network->hash_size);
	}

	if (pthread_rwlock_init(&network->successor_list_lock, NULL))
		return -1;

	for (i = 0; i < network->num_successors; i++)
		chord_nodecpy(network->successor_list + i, &network->local_node,
				network->hash_size);

	if (pthread_rwlock_init(&network->predecessor_lock, NULL))
		return -1;

	chord_nodecpy(&network->predecessor, &network->local_node,
			network->hash_size);
	network->pred_failed = 0;
	network->handler = options->handler;
	network->arg = options->arg;

	if (listen(network->ssocketfd, 10) == -1)
		return -1;

	return 0;
}

static void chord_pred_exit(struct chord_network *network)
{
	unsigned char command = PRED_EXIT;
	struct iovec write_buffer[] = {
		{
			.iov_base = &command,
			.iov_len = sizeof(command)
		},
		{
			.iov_base = network->predecessor.hash,
			.iov_len = network->hash_size
		},
		{
			.iov_base = &network->predecessor.address,
			.iov_len = sizeof(network->predecessor.address)
		},
		{
			.iov_base = &network->predecessor.port,
			.iov_len = sizeof(network->predecessor.port)
		},
		{
			.iov_base = &network->pred_failed,
			.iov_len = sizeof(network->pred_failed)
		}
	};
	int socketfd;

	socketfd = chord_connect(network->finger_table);

	if (socketfd != -1) {
		chord_writev(socketfd, write_buffer, sizeof(write_buffer) /
				sizeof(write_buffer[0]));

		close(socketfd);
	}
}

static void chord_succ_exit(struct chord_network *network)
{
	unsigned char command = SUCC_EXIT;
	int socketfd;

	socketfd = chord_connect(&network->predecessor);
	if (socketfd != -1) {
		chord_write(socketfd, &command, sizeof(command));
		close(socketfd);
	}
}

static void chord_exit(struct chord_network *network)
{
	chord_pred_exit(network);
	chord_succ_exit(network);
}

static void chord_cleanup(struct chord_network *network)
{
	int i;
	if (network) {
		if (network->finger_table)
			for (i = 0; i < network->hash_size * 8; i++)
				free((network->finger_table + i)->hash);

		free(network->finger_table_lock);
		free(network->finger_table);
		if (network->successor_list)
			for (i = 0; i < network->num_successors; i++)
				free((network->successor_list + i)->hash);

		free(network->successor_list);
		free(network->local_node.hash);
		free(network->predecessor.hash);
	}
	free(network);
}

static int chord_alloc(const struct chord_opts *options,
		struct chord_network **network)
{
	int i, retval = 0;

	*network = malloc(sizeof(struct chord_network));
	if (!*network) {
		retval = -1;
		goto out;
	}

	(*network)->hash_size = options->hash_size ? : HASH_SIZE;
	(*network)->num_successors =
			options->num_successors ? : NUM_SUCCESSORS;

	(*network)->finger_table = NULL;
	(*network)->successor_list = NULL;
	(*network)->local_node.hash = NULL;
	(*network)->predecessor.hash = NULL;
	(*network)->finger_table_lock = calloc((*network)->hash_size * 8,
			sizeof(*(*network)->finger_table_lock));

	if (!(*network)->finger_table_lock) {
		retval = -1;
		goto out;
	}
	(*network)->finger_table = calloc((*network)->hash_size * 8,
			sizeof(*(*network)->finger_table));

	if (!(*network)->finger_table) {
		retval = -1;
		goto out;
	}
	for (i = 1; i < (*network)->hash_size * 8; i++)
		((*network)->finger_table + i)->hash = NULL;

	for (i = 0; i < (*network)->hash_size * 8; i++) {
		((*network)->finger_table + i)->hash =
				malloc((*network)->hash_size);

		if (!((*network)->finger_table + i)->hash) {
			retval = -1;
			goto out;
		}
	}
	(*network)->successor_list = calloc((*network)->num_successors,
			sizeof(*(*network)->successor_list));

	if (!(*network)->successor_list) {
		retval = -1;
		goto out;
	}
	for (i = 1; i < (*network)->num_successors; i++)
		((*network)->successor_list + i)->hash = NULL;

	for (i = 0; i < (*network)->num_successors; i++) {
		((*network)->successor_list + i)->hash =
				malloc((*network)->hash_size);

		if (!((*network)->successor_list + i)->hash) {
			retval = -1;
			goto out;
		}
	}

	(*network)->local_node.hash = malloc((*network)->hash_size);
	if (!(*network)->local_node.hash) {
		retval = -1;
		goto out;
	}
	(*network)->predecessor.hash = malloc((*network)->hash_size);

	if (!(*network)->predecessor.hash)
		retval = -1;

	out:
	return retval;
}

int chord_successor_ext(const struct chord_node *node, size_t hash_size,
		const unsigned char *hash, struct chord_node *successor)
{
	unsigned char last_node;
	int retval, socketfd;

	socketfd = chord_connect(node);
	if (socketfd == -1)
		return -1;

	retval = chord_next_node(socketfd, hash, hash_size, successor,
			&last_node);

	close(socketfd);

	while (!last_node && retval != -1) {
		socketfd = chord_connect(successor);
		if (socketfd == -1)
			return -1;

		retval = chord_next_node(socketfd, hash, hash_size, successor,
				&last_node);

		close(socketfd);
	}

	return retval;
}

int chord_init_node(const char *node, const char *service, struct chord_node *chord_node)
{
	const struct addrinfo hints =  {
		.ai_family = AF_INET,
		.ai_socktype = SOCK_STREAM,
		.ai_protocol = IPPROTO_TCP,
		.ai_flags = 0,
		.ai_addrlen = 0,
		.ai_addr = NULL,
		.ai_canonname = NULL,
		.ai_next =  NULL
	};
	struct addrinfo *address;
	struct sockaddr_in *node_address;

	if (getaddrinfo(node, service, &hints, &address))
		return -1;

	node_address = (struct sockaddr_in *) address->ai_addr;

	chord_node->address = node_address->sin_addr.s_addr;
	chord_node->port = node_address->sin_port;

	freeaddrinfo(address);

	return 0;
}

int chord_oob(struct chord_node *node)
{
	int socketfd;
	int retval = 0;
	unsigned char command = OOB;

	socketfd = chord_connect(node);
	if (socketfd == -1)
		return -1;

	pthread_cleanup_push(chord_close, &socketfd);
	if (chord_write(socketfd, &command, sizeof(command)) == -1) {
		retval = -1;
		goto out;
	}
	retval = 0;

	out:
	pthread_cleanup_pop(0);
	if (retval == -1) {
		close(socketfd);
		return -1;
	}
	return socketfd;
}

int chord_successor(struct chord_network *network, unsigned char *hash,
		struct chord_node *successor)
{
	unsigned char last_node;
	int socketfd, retval = 0;

	last_node = find_next_node(network, hash, successor);

	while (!last_node && retval != -1) {
		socketfd = chord_connect(successor);
		if (socketfd == -1) {
			retval = -1;
			break;
		}
		pthread_cleanup_push(chord_close, &socketfd);
		retval = chord_next_node(socketfd, hash, network->hash_size,
				successor, &last_node);

		pthread_cleanup_pop(1);
	}
	return retval;
}

int chord_start(const struct chord_opts *options,
		struct chord_network **network)
{
	int retval;

	retval = chord_alloc(options, network);
	if (retval == -1)
		goto out;

	retval = chord_init(options, *network);
	if (retval == -1)
		goto out;

	if (options->join) {
		retval = chord_join(options, *network);
		if (retval == -1)
			goto out;
	}

	pthread_create(&(*network)->server_thread, NULL, chord_serve_request,
			*network);

	pthread_create(&(*network)->maint_thread, NULL, chord_maintainance,
			*network);

	out:
	if (retval == -1)
		chord_cleanup(*network);

	return retval;
}

void chord_stop(struct chord_network *network)
{
	pthread_cancel(network->maint_thread);
	pthread_join(network->maint_thread, NULL);

	pthread_cancel(network->server_thread);
	pthread_join(network->server_thread, NULL);

	close(network->ssocketfd);

	chord_exit(network);
	chord_cleanup(network);
}
