#include "chord_common.h"

#include <pthread.h>

#include <gmp.h>

#include <openssl/evp.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <error.h>

void chord_free(void *arg)
{
	void **ptr = arg;
	free(*ptr);
}

void chord_mpz_clear(void *arg)
{
	mpz_t *num = arg;
	mpz_clear(*num);
}

void chord_mutex_unlock(void *arg)
{
	pthread_mutex_t *lock = arg;
	pthread_mutex_unlock(lock);
}

void chord_rwlock_unlock(void *arg)
{
	pthread_rwlock_t *lock = arg;
	pthread_rwlock_unlock(lock);
}

int chord_nodecmp(const struct chord_node *node1,
		const struct chord_node *node2, size_t hash_size)
{
	int retval;

	retval = memcmp(node1->hash, node2->hash, hash_size);
	if (retval)
		goto out;

	if (node1->address != node2->address || node1->port != node2->port)
		retval = 1;

	out:
	return retval;
}

void chord_nodecpy(struct chord_node *dest, const struct chord_node *src,
		size_t hash_size)
{
	memcpy(dest->hash, src->hash, hash_size);
	dest->address = src->address;
	dest->port = src->port;
}

int chord_read(int fd, void *buf, size_t count)
{
	int retval = 0;
	ssize_t data_read;

	while (count) {
		data_read = read(fd, buf, count);

		if (data_read == -1 && errno == EINTR)
			continue;

		if (data_read == -1 || !data_read) {
			retval = -1;
			break;
		}

		count -= data_read;
		buf += data_read;
	}

	return retval;
}

int chord_write(int fd, void *buf, size_t count)
{
	int retval = 0;
	ssize_t data_wrote;

	while (count) {
		data_wrote = write(fd, buf, count);

		if (data_wrote == -1) {
			if (errno == EINTR)
				continue;

			retval = -1;
			break;
		}

		count -= data_wrote;
		buf += data_wrote;
	}

	return retval;
}

int chord_readv(int fd, struct iovec *iov, int iovcnt)
{
	int retval = 0;
	ssize_t data_read;

	while (iovcnt) {
		data_read = readv(fd, iov, iovcnt);

		if (data_read == -1 && errno == EINTR)
			continue;

		if (data_read == -1 || !data_read) {
			retval = -1;
			break;
		}

		while (data_read) {
			if (data_read >= iov->iov_len) {
				data_read -= iov->iov_len;
				iov++;
				iovcnt--;
			}
			else {
				iov->iov_len -= data_read;
				iov->iov_base += data_read;
				data_read = 0;
			}
		}
	}

	return retval;
}

int chord_writev(int fd, struct iovec *iov, int iovcnt)
{
	int retval = 0;
	ssize_t data_wrote;

	while (iovcnt) {
		data_wrote = writev(fd, iov, iovcnt);

		if (data_wrote == -1) {
			if (errno == EINTR)
				continue;
			retval = -1;
			break;
		}

		while (data_wrote) {
			if (data_wrote >= iov->iov_len) {
				data_wrote -= iov->iov_len;
				iov++;
				iovcnt--;
			}
			else {
				iov->iov_len -= data_wrote;
				iov->iov_base += data_wrote;
				data_wrote = 0;
			}
		}
	}

	return retval;
}

void chord_close(void *arg)
{
	int *socketfd = arg;

	close(*socketfd);
}

int chord_connect(const struct chord_node *node)
{
	int retval = 0;
	int socketfd;
	const struct sockaddr_in address = {
		.sin_family = AF_INET,
		.sin_port = node->port,
		.sin_addr = {
			.s_addr = node->address
		}
	};

	socketfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	retval = socketfd == -1 ? -1 : 0;
	if (!retval) {
		pthread_cleanup_push(chord_close, &socketfd);
		retval = connect(socketfd, (struct sockaddr *) &address,
				sizeof(address));

		pthread_cleanup_pop(retval == -1);
	}
	if (!retval)
		return socketfd;

	return retval;
}

int chord_check(struct chord_node *node)
{
	unsigned char command = CHECK;
	unsigned char node_status;
	int socketfd, retval = 0;

	socketfd = chord_connect(node);

	if (socketfd == -1)
		return -1;

	if (chord_write(socketfd, &command, sizeof(command)) == -1) {
		retval = -1;
		goto out;
	}

	if (chord_read(socketfd, &node_status, sizeof(node_status)) == -1) {
		retval = -1;
		goto out;
	}

	if (node_status)
		retval = -1;
	
	out:
	if (close(socketfd) == -1)
		error(EXIT_FAILURE, errno, "failed to close socket");

	return retval;
}

unsigned char find_next_node(struct chord_network *network,
		const unsigned char *hash, struct chord_node *next_node)
{
	mpz_t identifier, local_id, mod_amount, node_id, tmp;
	int i;
	unsigned char retval;

	mpz_init2(identifier, network->hash_size * 8);
	pthread_cleanup_push(chord_mpz_clear, &identifier);

	mpz_init2(local_id, network->hash_size * 8);
	pthread_cleanup_push(chord_mpz_clear, &local_id);

	mpz_init2(mod_amount, network->hash_size * 8 + 1);
	pthread_cleanup_push(chord_mpz_clear, &mod_amount);

	mpz_init2(node_id, network->hash_size * 8);
	pthread_cleanup_push(chord_mpz_clear, &node_id);

	mpz_init2(tmp, network->hash_size * 8);
	pthread_cleanup_push(chord_mpz_clear, &tmp);

	mpz_import(identifier, network->hash_size / sizeof(*hash), 1,
			sizeof(*hash), 1, 0, hash);

	pthread_rwlock_rdlock(&network->local_node_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->local_node_lock);
	mpz_import(local_id, network->hash_size /
			sizeof(*network->local_node.hash), 1,
			sizeof(*network->local_node.hash),
			1, 0, network->local_node.hash);

	pthread_cleanup_pop(1);

	mpz_ui_pow_ui(mod_amount, 2, network->hash_size * 8);

	mpz_sub(tmp, identifier, local_id);
	if (mpz_sgn(tmp) == -1)
		mpz_add(tmp, tmp, mod_amount);


	for (i = mpz_sizeinbase(tmp, 2) - 1; i != -1; i--) {
		pthread_rwlock_rdlock(network->finger_table_lock + i);
		pthread_cleanup_push(chord_rwlock_unlock,
				network->finger_table_lock + i);

		mpz_import(node_id, network->hash_size /
				sizeof(*network->finger_table->hash), 1,
				sizeof(*network->finger_table->hash), 1, 0,
				network->finger_table[i].hash);

		pthread_cleanup_pop(1);

		if (mpz_cmp(identifier, local_id) >= 0) {
			if (mpz_cmp(node_id, identifier) <= 0 &&
					mpz_cmp(node_id, local_id) > 0)
				break;
		}
		else
			if (mpz_cmp(node_id, identifier) <= 0 ||
					mpz_cmp(node_id, local_id) > 0)
				break;
	}

	if (i == -1) {
		pthread_rwlock_rdlock(network->finger_table_lock);
		pthread_cleanup_push(chord_rwlock_unlock,
				network->finger_table_lock);

		chord_nodecpy(next_node, network->finger_table,
				network->hash_size);

		pthread_cleanup_pop(1);
		retval = 1;
	}
	else {
		pthread_rwlock_rdlock(network->finger_table_lock + i);
		pthread_cleanup_push(chord_rwlock_unlock,
				network->finger_table_lock + i);

		chord_nodecpy(next_node, network->finger_table + i,
				network->hash_size);

		pthread_cleanup_pop(1);
		retval = 0;
	}

	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);

	return retval;
}

int chord_next_node(int socketfd, const unsigned char *hash, size_t hash_size,
		struct chord_node *next_node, unsigned char *last_node)
{
	int retval = 0;
	unsigned char command = CHORD_FIND;
	struct iovec write_buffer[] = {
		{
			.iov_base = &command,
			.iov_len = sizeof(command)
		},
		{
			.iov_base = (void *) hash,
			.iov_len = hash_size
		}
	};
	struct iovec read_buffer[] = {
		{
			.iov_base = next_node->hash,
			.iov_len = hash_size
		},
		{
			.iov_base = &next_node->address,
			.iov_len = sizeof(next_node->address)
		},
		{
			.iov_base = &next_node->port,
			.iov_len = sizeof(next_node->port)
		},
		{
			.iov_base = last_node,
			.iov_len = sizeof(*last_node)
		}
	};

	retval = chord_writev(socketfd, write_buffer,
			sizeof(write_buffer) / sizeof(write_buffer[0]));

	retval = retval ? : chord_readv(socketfd, read_buffer,
			sizeof(read_buffer) / sizeof(read_buffer[0]));

	return retval;
}

void successor_failed(struct chord_network *network)
{
	int i, retval;

	pthread_rwlock_wrlock(&network->successor_list_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->successor_list_lock);

	pthread_rwlock_wrlock(network->finger_table_lock);
	pthread_cleanup_push(chord_rwlock_unlock, network->finger_table_lock);

	retval = chord_check(network->finger_table);

	if (retval == -1)
		chord_nodecpy(network->finger_table, network->successor_list,
				network->hash_size);

	pthread_cleanup_pop(1);

	if (retval != -1)
		goto out;

	for (i = 0; i < network->num_successors - 1; i++)
		chord_nodecpy(network->successor_list + i,
				network->successor_list + i + 1,
				network->hash_size);

	pthread_rwlock_rdlock(&network->local_node_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->local_node_lock);
	chord_nodecpy(network->successor_list + network->num_successors - 1,
			&network->local_node, network->hash_size);

	pthread_cleanup_pop(1);

	out:
	pthread_cleanup_pop(1);
}

void hash_address(struct chord_network *network)
{
	EVP_MD_CTX ctx;
	unsigned int hash_len;
	unsigned char hash[EVP_MAX_MD_SIZE];

	EVP_MD_CTX_init(&ctx);
	EVP_DigestInit_ex(&ctx, EVP_sha1(), NULL);
	EVP_DigestUpdate(&ctx, &network->local_node.address,
			sizeof(network->local_node.address));
	EVP_DigestFinal_ex(&ctx, hash, &hash_len);
	EVP_MD_CTX_cleanup(&ctx);

	if (hash_len < network->hash_size) {
		memcpy(network->local_node.hash, hash, hash_len);
		memset(network->local_node.hash + hash_len, 0,
				network->hash_size - hash_len);
	}
	else
		memcpy(network->local_node.hash, hash, network->hash_size);
}

unsigned char hash_inrange(unsigned char *hash, unsigned char *bot_hash,
		unsigned char *top_hash, size_t hash_size)
{
	unsigned char retval = 0;
	mpz_t top_id, bot_id, id;

	mpz_init2(top_id, hash_size * 8);
	mpz_init2(bot_id, hash_size * 8);
	mpz_init2(id, hash_size * 8);

	mpz_import(top_id, hash_size / sizeof(*top_hash), 1, sizeof(*top_hash),
			1, 0, top_hash);
	mpz_import(bot_id, hash_size / sizeof(*bot_hash), 1, sizeof(*bot_hash),
			1, 0, bot_hash);
	mpz_import(id, hash_size / sizeof(*hash), 1, sizeof(*hash), 1, 0, hash);

	if (mpz_cmp(top_id, bot_id) > 0) {
		if (mpz_cmp(top_id, id) > 0 && mpz_cmp(bot_id, id) < 0)
			retval = 1;
	}
	else
		if (mpz_cmp(top_id, id) > 0 || mpz_cmp(bot_id, id) < 0)
			retval = 1;

	mpz_clear(top_id);
	mpz_clear(bot_id);
	mpz_clear(id);

	return retval;
}

int chord_get_successors(struct chord_network *network,
		struct chord_node_list *list)
{
	int retval = 0;

	retval = pthread_rwlock_rdlock(&network->successor_list_lock);
	if (!retval) {
		pthread_cleanup_push(chord_rwlock_unlock,
				&network->successor_list_lock);

		list->num_nodes = network->num_successors;
		list->nodes = calloc(list->num_nodes, sizeof(*list->nodes));
		retval = list->nodes ? 0 : -1;
		if (!retval) {
			pthread_cleanup_push(free, list->nodes);
			memcpy(list->nodes, network->successor_list,
					list->num_nodes *
					sizeof(*list->nodes));

			pthread_cleanup_pop(retval == -1);
		}
		pthread_cleanup_pop(1);
	}

	return retval;
}

void chord_node_list_destroy(struct chord_node_list *list)
{
	free(list->nodes);
}
