#include "chord_common.h"

#include <pthread.h>

#include <gmp.h>

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <error.h>

#define SLEEP_TIME 1

static int chord_get_successor(int socketfd, struct chord_node *next_node,
		size_t hash_size)
{
	unsigned char command = GET_SUCCESSOR;
	struct iovec read_buffer[] = {
		{
			.iov_base = next_node->hash,
			.iov_len = hash_size
		},
		{
			.iov_base = &next_node->address,
			.iov_len = sizeof(next_node->address)
		},
		{
			.iov_base = &next_node->port,
			.iov_len = sizeof(next_node->port)
		}
	};
	int retval;

	retval = chord_write(socketfd, &command, sizeof(command));
	if (retval == -1)
		goto out;

	retval = chord_readv(socketfd, read_buffer,
			sizeof(read_buffer) / sizeof(read_buffer[0]));

	out:
	return retval;
}


static int chord_fix_successors(struct chord_network *network)
{
	int i, socketfd, retval = 0;
	struct chord_node next_node;

	next_node.hash = NULL;
	pthread_cleanup_push(chord_free, &next_node.hash);

	next_node.hash = malloc(network->hash_size);
	if (!next_node.hash) {
		retval = -1;
		goto out;
	}

	pthread_rwlock_rdlock(network->finger_table_lock);
	pthread_cleanup_push(chord_rwlock_unlock, network->finger_table_lock);
	chord_nodecpy(&next_node, network->finger_table, network->hash_size);
	pthread_cleanup_pop(1);

	pthread_rwlock_wrlock(&network->successor_list_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->successor_list_lock);
	pthread_rwlock_rdlock(&network->local_node_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->local_node_lock);
	for (i = 0; i < network->num_successors; i++) {
		socketfd = chord_connect(&next_node);
		if (socketfd == -1) {
			retval = -1;
			break;
		}
		retval = chord_get_successor(socketfd, &next_node,
				network->hash_size);

		/* TODO: don't check close value */
		/* TODO: remove error.h ? */
		if (close(socketfd) == -1)
			error(EXIT_FAILURE, errno, "could not close socket");

		if (retval == -1)
			break;

		if (chord_nodecmp(&next_node, &network->local_node,
				network->hash_size)) {

			chord_nodecpy(network->successor_list + i, &next_node,
					network->hash_size);
			continue;
		}
		do {
			chord_nodecpy(network->successor_list + i,
					&network->local_node,
					network->hash_size);
		} while (++i < network->num_successors);
		break;
	}
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);

	out:
	pthread_cleanup_pop(1);
	return retval;
}
		

static void chord_fix_fingers(struct chord_network *network)
{
	int i;
	mpz_t local_id, finger_id, mod_amount;
	unsigned char *finger_hash;
	struct chord_node finger_entry;
	size_t hash_size;

	finger_hash = NULL;
	finger_entry.hash = NULL;

	pthread_cleanup_push(chord_free, &finger_hash);
	pthread_cleanup_push(chord_free, &finger_entry.hash);

	finger_hash = malloc(network->hash_size);
	if (!finger_hash)
		goto out;

	finger_entry.hash = malloc(network->hash_size);
	if (!finger_entry.hash)
		goto out;

	mpz_init2(local_id, network->hash_size * 8);
	pthread_cleanup_push(chord_mpz_clear, &local_id);

	mpz_init2(finger_id, network->hash_size * 8);
	pthread_cleanup_push(chord_mpz_clear, &finger_id);

	mpz_init2(mod_amount, network->hash_size * 8 + 1);
	pthread_cleanup_push(chord_mpz_clear, &mod_amount);

	mpz_ui_pow_ui(mod_amount, 2, network->hash_size * 8);

	pthread_rwlock_rdlock(&network->local_node_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->local_node_lock);
	mpz_import(local_id, network->hash_size /
			sizeof(*network->local_node.hash), 1,
			sizeof(*network->local_node.hash), 1, 0,
			network->local_node.hash);

	pthread_cleanup_pop(1);

	for (i = 1; i < network->hash_size * 8; i++) {
		mpz_ui_pow_ui(finger_id, 2, i);
		mpz_add(finger_id, finger_id, local_id);
		if (mpz_cmp(finger_id, mod_amount) >= 0)
			mpz_sub(finger_id, finger_id, mod_amount);

		mpz_export(finger_hash, &hash_size, 1,
				sizeof(*finger_hash), 1, 0, finger_id);

		if (network->hash_size - hash_size) {
			memmove(finger_hash + network->hash_size - hash_size,
					finger_hash, hash_size);
			memset(finger_hash, 0, network->hash_size - hash_size);
		}

		if (chord_successor(network, finger_hash, &finger_entry) == -1)
			continue;

		pthread_rwlock_wrlock(network->finger_table_lock + i);
		pthread_cleanup_push(chord_rwlock_unlock,
				network->finger_table_lock + i);

		chord_nodecpy(network->finger_table + i, &finger_entry,
				network->hash_size);

		pthread_cleanup_pop(1);
	}

	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);

	out:
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
}

static int chord_get_spred(struct chord_network *network,
		struct chord_node *spred, unsigned char *spred_failed)
{
	int socketfd, retval;
	unsigned char command = GET_PREDECESSOR;

	while (1) {
		struct iovec read_buffer[] = {
			{
				.iov_base = spred->hash,
				.iov_len = network->hash_size
			},
			{
				.iov_base = &spred->address,
				.iov_len = sizeof(spred->address)
			},
			{
				.iov_base = &spred->port,
				.iov_len = sizeof(spred->port)
			},
			{
				.iov_base = spred_failed,
				.iov_len = sizeof(*spred_failed)
			}
		};

		pthread_rwlock_rdlock(network->finger_table_lock);
		pthread_cleanup_push(chord_rwlock_unlock,
				network->finger_table_lock);

		socketfd = chord_connect(network->finger_table);
		pthread_cleanup_pop(1);

		if (socketfd == -1) {
			successor_failed(network);
			continue;
		}
		pthread_cleanup_push(chord_close, &socketfd);

		retval = chord_write(socketfd, &command, sizeof(command));
		retval = retval ? : chord_readv(socketfd, read_buffer,
				sizeof(read_buffer) / sizeof(read_buffer[0]));

		pthread_cleanup_pop(1);

		if (retval != -1)
			break;

		successor_failed(network);
	}

	return retval;
}

static int chord_notify(struct chord_network *network)
{
	unsigned char command = NOTIFY;
	struct iovec write_buffer[] = {
		{
			.iov_base = &command,
			.iov_len = sizeof(command)
		},
		{
			.iov_base = network->local_node.hash,
			.iov_len = network->hash_size
		},
		{
			.iov_base = &network->local_node.address,
			.iov_len = sizeof(network->local_node.address)
		},
		{
			.iov_base = &network->local_node.port,
			.iov_len = sizeof(network->local_node.port)
		}
	};
	int socketfd, retval;

	/* TODO: limit retries */
	while (1) {
		pthread_rwlock_rdlock(network->finger_table_lock);
		pthread_cleanup_push(chord_rwlock_unlock,
				network->finger_table_lock);

		socketfd = chord_connect(network->finger_table);
		pthread_cleanup_pop(1);

		if (socketfd == -1) {
			successor_failed(network);
			continue;
		}
		pthread_cleanup_push(chord_close, &socketfd);
		pthread_rwlock_rdlock(&network->local_node_lock);
		pthread_cleanup_push(chord_rwlock_unlock,
				&network->local_node_lock);

		retval = chord_writev(socketfd, write_buffer,
				sizeof(write_buffer) / sizeof(write_buffer[0]));

		pthread_cleanup_pop(1);
		pthread_cleanup_pop(1);

		if (retval != -1)
			break;

		successor_failed(network);
	}
	return retval;
}

static int chord_stabilize(struct chord_network *network)
{
	int retval;
	struct chord_node spred;
	unsigned char spred_invalid;

	spred.hash = NULL;
	pthread_cleanup_push(chord_free, &spred.hash);

	spred.hash = malloc(network->hash_size);
	if (!spred.hash) {
		retval = -1;
		goto out;
	}
	retval = chord_get_spred(network, &spred, &spred_invalid);
	if (!spred_invalid && retval != -1) {
		pthread_rwlock_rdlock(network->finger_table_lock);
		pthread_cleanup_push(chord_rwlock_unlock,
				network->finger_table_lock);

		pthread_rwlock_rdlock(&network->local_node_lock);
		pthread_cleanup_push(chord_rwlock_unlock,
				&network->local_node_lock);

		spred_invalid = !hash_inrange(spred.hash,
				network->local_node.hash,
				network->finger_table->hash,
				network->hash_size);

		pthread_cleanup_pop(1);
		pthread_cleanup_pop(1);

		if (!spred_invalid) {
			pthread_rwlock_wrlock(network->finger_table_lock);
			pthread_cleanup_push(chord_rwlock_unlock,
					network->finger_table_lock);

			chord_nodecpy(network->finger_table, &spred,
					network->hash_size);

			pthread_cleanup_pop(1);
		}
	}
	out:
	pthread_cleanup_pop(1);
	return retval;
}

static void chord_check_pred(struct chord_network *network)
{
	pthread_rwlock_rdlock(&network->predecessor_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->predecessor_lock);

	if (chord_check(&network->predecessor) == -1)
		network->pred_failed = 1;
	else
		network->pred_failed = 0;

	pthread_cleanup_pop(1);
}

void *chord_maintainance(void *arg)
{
	struct chord_network *network = arg;

	while (1) {
		chord_check_pred(network);
		chord_stabilize(network);
		chord_notify(network);
		chord_fix_successors(network);
		chord_fix_fingers(network);

		/* TODO: sleep properly */
		sleep(SLEEP_TIME);
	}
}
