#include "chord_common.h"

#include <pthread.h>

#include <stdlib.h>
#include <error.h>

static void chord_serve_check(int socketfd)
{
	unsigned char retval = 0;

	chord_write(socketfd, &retval, sizeof(&retval));
}

static void chord_serve_find(struct chord_network *network, int socketfd)
{
	unsigned char last_node, *hash;
	struct chord_node next_node;
	struct iovec write_buffer[] = {
		{
			.iov_base = NULL,
			.iov_len = network->hash_size
		},
		{
			.iov_base = &next_node.address,
			.iov_len = sizeof(next_node.address)
		},
		{
			.iov_base = &next_node.port,
			.iov_len = sizeof(next_node.port)
		},
		{
			.iov_base = &last_node,
			.iov_len = sizeof(last_node)
		}
	};

	hash = NULL;
	next_node.hash = NULL;
	pthread_cleanup_push(chord_free, &hash);
	pthread_cleanup_push(chord_free, &next_node.hash);

	hash = malloc(network->hash_size);
	if (!hash)
		goto out;

	next_node.hash = malloc(network->hash_size);
	if (!next_node.hash)
		goto out;

	write_buffer->iov_base = next_node.hash;

	if (chord_read(socketfd, hash, network->hash_size) == -1)
		goto out;

	last_node = find_next_node(network, hash, &next_node);

	chord_writev(socketfd, write_buffer,
			sizeof(write_buffer) / sizeof(write_buffer[0]));

	out:
	pthread_cleanup_pop(1);
	pthread_cleanup_pop(1);
}

static void chord_serve_pred(struct chord_network *network, int socketfd)
{
	struct iovec write_buffer[] = {
		{
			.iov_base = network->predecessor.hash,
			.iov_len = network->hash_size
		},
		{
			.iov_base = &network->predecessor.address,
			.iov_len = sizeof(network->predecessor.address)
		},
		{
			.iov_base = &network->predecessor.port,
			.iov_len = sizeof(network->predecessor.port)
		},
		{
			.iov_base = &network->pred_failed,
			.iov_len = sizeof(network->pred_failed)
		}
	};

	pthread_rwlock_rdlock(&network->predecessor_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->predecessor_lock);
	chord_writev(socketfd, write_buffer,
			sizeof(write_buffer) / sizeof(write_buffer[0]));

	pthread_cleanup_pop(1);
}

/* TODO: read into temporary node */
static void chord_serve_notify(struct chord_network *network, int socketfd)
{
	struct chord_node new_pred;
	struct iovec read_buffer[] = {
		{
			.iov_base = NULL,
			.iov_len = network->hash_size
		},
		{
			.iov_base = &new_pred.address,
			.iov_len = sizeof(network->predecessor.address)
		},
		{
			.iov_base = &new_pred.port,
			.iov_len = sizeof(network->predecessor.port)
		}
	};
	int retval;

	new_pred.hash = NULL;
	pthread_cleanup_push(chord_free, &new_pred.hash);

	new_pred.hash = malloc(network->hash_size);
	if (!new_pred.hash)
		goto out;

	read_buffer->iov_base = new_pred.hash;

	retval = chord_readv(socketfd, read_buffer,
			sizeof(read_buffer) / sizeof(read_buffer[0]));

	if (retval == -1)
		goto out;

	if (!network->pred_failed) {
		pthread_rwlock_rdlock(&network->predecessor_lock);
		pthread_cleanup_push(chord_rwlock_unlock,
				&network->predecessor_lock);

		pthread_rwlock_rdlock(&network->local_node_lock);
		pthread_cleanup_push(chord_rwlock_unlock,
				&network->local_node_lock);

		retval = hash_inrange(new_pred.hash, network->predecessor.hash,
				network->local_node.hash, network->hash_size);

		pthread_cleanup_pop(1);
		pthread_cleanup_pop(1);

		if (!retval)
			goto out;
	}

	pthread_rwlock_wrlock(&network->predecessor_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->predecessor_lock);
	chord_nodecpy(&network->predecessor, &new_pred, network->hash_size);
	pthread_cleanup_pop(1);
	network->pred_failed = 0;

	out:
	pthread_cleanup_pop(1);
}

static void chord_serve_succ(struct chord_network *network, int socketfd)
{
	struct iovec write_buffer[] = {
		{
			.iov_base = network->finger_table->hash,
			.iov_len = network->hash_size
		},
		{
			.iov_base = &network->finger_table->address,
			.iov_len = sizeof(network->finger_table->address)
		},
		{
			.iov_base = &network->finger_table->port,
			.iov_len = sizeof(network->finger_table->port)
		}
	};

	pthread_rwlock_rdlock(network->finger_table_lock);
	pthread_cleanup_push(chord_rwlock_unlock, network->finger_table_lock);
	chord_writev(socketfd, write_buffer,
			sizeof(write_buffer) / sizeof(write_buffer[0]));

	pthread_cleanup_pop(1);
}

static void chord_serve_pred_failed(struct chord_network *network, int socketfd)
{
	struct chord_node new_pred;
	unsigned char new_pred_failed;
	struct iovec read_buffer[] = {
		{
			.iov_base = NULL,
			.iov_len = network->hash_size
		},
		{
			.iov_base = &new_pred.address,
			.iov_len = sizeof(new_pred.address)
		},
		{
			.iov_base = &new_pred.port,
			.iov_len = sizeof(new_pred.port)
		},
		{
			.iov_base = &new_pred_failed,
			.iov_len = sizeof(new_pred_failed)
		}
	};
	new_pred.hash = NULL;
	pthread_cleanup_push(chord_free, &new_pred.hash);

	new_pred.hash = malloc(network->hash_size);
	if (!new_pred.hash)
		goto out;

	read_buffer[0].iov_base = new_pred.hash;

	if (chord_readv(socketfd, read_buffer, sizeof(read_buffer) /
			sizeof(read_buffer[0])) == -1)
		goto out;

	pthread_rwlock_wrlock(&network->predecessor_lock);
	pthread_cleanup_push(chord_rwlock_unlock, &network->predecessor_lock);
	chord_nodecpy(&network->predecessor, &new_pred, network->hash_size);
	network->pred_failed = new_pred_failed;
	pthread_cleanup_pop(1);

	out:
	pthread_cleanup_pop(1);
}

static void chord_serve_succ_failed(struct chord_network *network)
{
	successor_failed(network);
}

void *chord_serve_request(void *arg)
{
	struct chord_network *network = arg;
	struct sockaddr_storage peer_addr;
	int peer_sfd;
	socklen_t peer_addr_len;
	unsigned char command;

	while (1) {
		peer_addr_len = sizeof(peer_addr);

		peer_sfd = accept(network->ssocketfd,
				(struct sockaddr *) &peer_addr, &peer_addr_len);

		if (peer_sfd == -1)
			continue;

		pthread_cleanup_push(chord_close, &peer_sfd);

		if (chord_read(peer_sfd, &command, sizeof(command)) == -1)
			goto out;

		switch (command) {
		case CHORD_FIND:
			chord_serve_find(network, peer_sfd);
			break;
		case GET_PREDECESSOR:
			chord_serve_pred(network, peer_sfd);
			break;
		case NOTIFY:
			chord_serve_notify(network, peer_sfd);
			break;
		case GET_SUCCESSOR:
			chord_serve_succ(network, peer_sfd);
			break;
		case CHECK:
			chord_serve_check(peer_sfd);
			break;
		case PRED_EXIT:
			chord_serve_pred_failed(network, peer_sfd);
			break;
		case SUCC_EXIT:
			chord_serve_succ_failed(network);
			break;
		case OOB:
			if (network->handler) {
				network->handler(peer_sfd, network->arg);
				break;
			}
		default:
			error(0, 0, "invalid command received on socket");
		}

		out:
		pthread_cleanup_pop(1);
	}
}
