#include <chord.h>

#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>

#define SLEEP_TIME 3
#define NUM_NODES 5
#define HASH_SIZE 20
#define SERVICE "54321"

struct options {
	int num_nodes;
	char *service;
	int sleep_time;
	int hash_size;
};

static int check_consistency(struct chord_node *nodes,
		struct chord_network **networks, struct options *opts)
{
	struct chord_node node;
	int i, j;

	node.hash = calloc(opts->hash_size, sizeof(*node.hash));
	if (!node.hash)
		return -1;

	for (i = 0; i < opts->num_nodes; i++) {
		for (j = 0; j < opts->num_nodes; j++) {
			if (chord_successor(*(networks + i), (nodes + j)->hash,
					&node) == -1) {

				printf("lookup failed\n");
				return -1;
			}

			if (memcmp(node.hash, (nodes + (j + 1) %
					opts->num_nodes)->hash,
					opts->hash_size)) {

				printf("network inconsistent\n");
				return -1;
			}
		}
	}
	printf("network consistent\n");
	return 0;
}

static int print_network(struct chord_node *nodes, struct options *opts)
{
	int i, j;
	struct in_addr address;
	char paddress[INET_ADDRSTRLEN];

	for (i = 0; i < opts->num_nodes; i ++) {
		address.s_addr = (nodes + i)->address;
		if (!inet_ntop(AF_INET, &address, paddress, sizeof(paddress)))
			return -1;

		printf("address: %s\n", paddress);
		printf("hash: ");
		for (j = 0; j < opts->hash_size; j++)
			printf("%.2hhx", *((nodes + i)->hash + j));

		printf("\n");
	}
	
	return 0;
}

static int discover_network(struct chord_node *nodes,
		struct chord_network **networks, struct options *opts)
{
	int i;
	unsigned char *hash;

	hash = calloc(opts->hash_size, sizeof(*hash));
	memset(hash, 0, opts->hash_size * sizeof(*hash));

	if (chord_successor(*networks, hash, nodes) == -1)
		return -1;

	for (i = 1; i < opts->num_nodes; i++)
		if (chord_successor(*networks, (nodes + i - 1)->hash,
			nodes + i) == -1)

			return -1;

	return 0;
}

static int parse_options(int argc, char *argv[], struct options *opts)
{
	const struct option longopts[] = {
		{
			.name = "nodes",
			.has_arg = required_argument,
			.flag = NULL,
			.val = 'n'
		},
		{
			.name = "port",
			.has_arg = required_argument,
			.flag = NULL,
			.val = 'p'
		},
		{
			.name = "wait",
			.has_arg = required_argument,
			.flag = NULL,
			.val = 'w'
		},
		{
			.name = "hashsize",
			.has_arg = required_argument,
			.flag = NULL,
			.val = 's'
		},
		{
			.name = NULL,
			.has_arg = 0,
			.flag = NULL,
			.val = 0
		}
	};
	const char optstring[] = "n:p:w:s:";

	while (1) {
		switch (getopt_long(argc, argv, optstring, longopts, NULL)) {
			case -1:
				return 0;
			case 0:
				printf("incorrect usage\n");
				return -1;
			case 'n':
				opts->num_nodes = atoi(optarg);
				if (opts->num_nodes <= 0)
					return -1;
				break;
			case 'p':
				opts->service = optarg;
				break;
			case 'w':
				opts->sleep_time = atoi(optarg);
				if (opts->sleep_time < 0)
					return -1;
				break;
			case 's':
				opts->hash_size = atoi(optarg);
				if (opts->hash_size <= 0)
					return -1;
				break;
		}
	}
}

int main(int argc, char *argv[])
{
	struct chord_network **networks;
	struct chord_node *nodes;
	int i;
	struct chord_opts options;
	char local_node[INET_ADDRSTRLEN];
	struct options opts;

	opts.num_nodes = NUM_NODES;
	opts.service = SERVICE;
	opts.sleep_time = SLEEP_TIME;
	opts.hash_size = HASH_SIZE;

	if (parse_options(argc, argv, &opts) == -1) {
		printf("incorrect usage\n");
		exit(EXIT_FAILURE);
	}

	options.join = 0;
	options.local_node = local_node;
	options.local_service = opts.service;
	options.node = "127.0.0.1";
	options.service = opts.service;
	options.num_successors = 0;
	options.hash_size = opts.hash_size;

	networks = calloc(opts.num_nodes, sizeof(*networks));
	if (!networks)
		exit(EXIT_FAILURE);

	nodes = calloc(opts.num_nodes, sizeof(*nodes));
	if (!nodes)
		exit(EXIT_FAILURE);

	for (i = 0; i < opts.num_nodes; i++) {
		(nodes + i)->hash = calloc(opts.hash_size,
				sizeof(*nodes->hash));
		if (!(nodes + i)->hash)
			exit(EXIT_FAILURE);
	}

	strcpy(local_node, "127.0.0.1");
	if (chord_start(&options, networks) == -1)
		exit(EXIT_FAILURE);

	for (i = 1, options.join = 1; i < opts.num_nodes; i++) {
		sprintf(local_node, "127.0.0.%d", i + 1);
		if (chord_start(&options, networks + i) == -1)
			exit(EXIT_FAILURE);
	}

	sleep(opts.sleep_time);

	if (discover_network(nodes, networks, &opts) == -1)
		exit(EXIT_FAILURE);
	if (print_network(nodes, &opts) == -1)
		exit(EXIT_FAILURE);
	if (check_consistency(nodes, networks, &opts) == -1)
		exit(EXIT_FAILURE);

	exit(EXIT_SUCCESS);
}
