#include <galois.h>

#include <p2pvdn/socket.h>
#include <p2pvdn/file_opts.h>
#include <p2pvdn/hash.h>
#include <p2pvdn/peer.h>
#include <p2pvdn/packet.h>

#include <chord.h>

#include <openssl/rand.h>

#include <fstream>
#include <vector>
#include <set>
#include <map>
#include <algorithm>
#include <functional>

#include <getopt.h>
#include <sys/stat.h>

namespace disperse {
	struct opts {
		std::string input_path;
		unsigned num_slices;
		unsigned packet_size;
		unsigned coeff_size;
		unsigned key_size;
		unsigned num_peers;
		unsigned peer_slices;
		std::string node;
		std::string service;

		opts(int argc, char * const argv[])
			: num_slices(20), packet_size(2048), coeff_size(1), key_size(20),
			num_peers(8), peer_slices(5), node("127.0.0.1"), service("50000")
		{
			const struct option opts[] = {
				{ "address", required_argument, NULL, 'a' },
				{ "port", required_argument, NULL, 'p' },
				{ "packet_size", required_argument, NULL, 's' },
				{ "coefficient_size", required_argument, NULL,'c' },
				{ "num_slices", required_argument, NULL, 'n' },
				{ "num_peers", required_argument, NULL, 'P' },
				{ "peer_slices", required_argument, NULL, 'S' },
				{ NULL, 0, NULL, 0 }
			};
			const char *short_opts = "a:p:s:c:n:P:S:";
			int opt;

			while ((opt = getopt_long(argc, argv, short_opts, opts, NULL)) != -1) {
				switch (opt) {
					case 'a':
						node = optarg;
						break;
					case 'p':
						service = optarg;
						break;
					case 's':
						packet_size = atoi(optarg);
						break;
					case 'c':
						coeff_size = atoi(optarg);
						break;
					case 'n':
						num_slices = atoi(optarg);
						break;
					case 'P':
						num_peers = atoi(optarg);
						break;
					case 'S':
						peer_slices = atoi(optarg);
						break;
					default:
						throw std::invalid_argument("invalid argument");
				}
			}
			if (*(argv + optind))
				input_path = *(argv + optind);
			else
				throw std::invalid_argument("missing input file path");
		}
	};

	struct ctx {
		std::ifstream input;
		p2pvdn::file_opts file_opts;
		p2pvdn::hash hash;
		chord_node gw_node;
		std::vector<std::pair<p2pvdn::node, unsigned>> peers;

		ctx(const opts &opts);
	};

	std::vector<std::pair<p2pvdn::node, unsigned>> choose_peers(const opts &opts, const ctx &ctx)
	{
		std::map<p2pvdn::node, unsigned> peers;
		std::vector<unsigned char> key(ctx.hash.size());
		chord_node tmp_node;
		// TODO: chord::node class
		tmp_node.hash = new unsigned char[key.size()];

		for (unsigned peer_num = 0; peer_num < opts.num_peers; peer_num++) {
			RAND_pseudo_bytes(&key[0], key.size());
			chord_successor_ext(&ctx.gw_node, key.size(), &key[0], &tmp_node);
			peers[p2pvdn::node(tmp_node.address, tmp_node.port)] += opts.peer_slices;
		}
		delete[] tmp_node.hash;

		return std::vector<std::pair<p2pvdn::node, unsigned>>(peers.begin(), peers.end());
	}

	std::vector<p2pvdn::socket> start_disperse(
			const std::vector<std::pair<p2pvdn::node, unsigned>> &peers,
			const p2pvdn::hash &hash, const p2pvdn::file_opts &opts)
	{
		std::vector<p2pvdn::socket> sockets(peers.size());
		auto socket = sockets.begin();
		for (auto peer = peers.begin(); peer != peers.end(); peer++, socket++) {
			chord_node node;
			node.address = peer->first.get_address();
			node.port = peer->first.get_port();
			*socket = {chord_oob(&node)};
			*socket << p2pvdn::command::disperse << hash << opts << peer->second;
		}

		return sockets;
	}

	template<unsigned coefficient_size> static void disperse_file(ctx &ctx)
	{
		typedef typename p2pvdn::packet_type<coefficient_size>::value type;

		std::vector<p2pvdn::socket> sockets = start_disperse(ctx.peers, ctx.hash, ctx.file_opts);
		std::vector<std::vector<type>> slices(ctx.file_opts.num_slices,
				std::vector<type>(ctx.file_opts.packet_elements()));

		std::vector<type> coefficients(ctx.file_opts.num_slices);
		for (unsigned data_rem = ctx.file_opts.filesize; data_rem; ) {
			for (auto slice = slices.begin(); slice != slices.end(); slice++) {
				if (data_rem < ctx.file_opts.packet_size) {
					ctx.input.read(reinterpret_cast<char *>(&(*slice)[0]), data_rem);
					data_rem = 0;
				}
				else {
					ctx.input.read(reinterpret_cast<char *>(&(*slice)[0]),
							ctx.file_opts.packet_size);

					data_rem -= ctx.file_opts.packet_size;
				}
			}
			auto socket = sockets.begin();
			for (auto peer = ctx.peers.begin(); peer != ctx.peers.end(); peer++, socket++) {
				for (unsigned slice_num = 0; slice_num < peer->second; slice_num++) {
					RAND_pseudo_bytes(reinterpret_cast<unsigned char *>(&coefficients[0]),
							ctx.file_opts.header_size());

					auto packet = netcodec::encode<coefficient_size>(coefficients, slices);
					*socket << p2pvdn::packet<coefficient_size>(coefficients, std::move(packet));
				}
			}
		}

		for (auto socket = sockets.begin(); socket != sockets.end(); socket++)
			*socket << p2pvdn::command::close;
	}

	ctx::ctx(const opts &opts)
		: input(opts.input_path.c_str()), hash(opts.input_path, opts.key_size)
	{
		struct stat stats;
		stat(opts.input_path.c_str(), &stats);
		file_opts.filesize = stats.st_size;
		file_opts.num_slices = opts.num_slices;
		file_opts.packet_size = opts.packet_size;
		file_opts.coeff_size = opts.coeff_size;
		chord_init_node(opts.node.c_str(), opts.service.c_str(), &gw_node);
		peers = choose_peers(opts, *this);
	}
}

static void send_peers(const disperse::ctx &ctx)
{
	chord_node node;
	node.hash = new unsigned char[ctx.hash.size()];
	chord_successor_ext(&ctx.gw_node, ctx.hash.size(), ctx.hash.get(), &node);
	p2pvdn::socket socket(chord_oob(&node));
	delete[] node.hash;
	socket << p2pvdn::command::get_peer_list << ctx.hash << ctx.peers;
	socket << p2pvdn::command::close;
}

int main(int argc, char *argv[])
{
	disperse::opts opts(argc, argv);
	disperse::ctx ctx(opts);
	if (ctx.file_opts.coeff_size == 1)
		disperse::disperse_file<1>(ctx);
	else
		disperse::disperse_file<2>(ctx);

	send_peers(ctx);

	return EXIT_SUCCESS;
}
