#ifndef SOCKET_H
#define SOCKET_H

#include <exception>
#include <stdexcept>
#include <iostream>
#include <vector>
#include <functional>

#include <cerrno>

#include <unistd.h>
#include <arpa/inet.h>

namespace p2pvdn {
	enum class command : unsigned char {
		disperse, stream, key_info, get_peer_list, send_peer_list, close
	};

	class network_error : public std::exception {
	};

	class connection_closed : public network_error {
	};

	class invalid_socket : public network_error {
	};

	class file_descriptor_limit_reached : public network_error {
	};

	class socket;

	const socket &operator<<(const socket &, unsigned);
	const socket &operator>>(const socket &, unsigned &);
	const socket &operator<<(const socket &, unsigned char);
	const socket &operator>>(const socket &, unsigned char &);
	template<class type> const socket &operator<<(const socket &, const std::vector<type> &);
	template<class type> const socket &operator>>(const socket &, std::vector<type> &);
	template<class first_type, class second_type>
		const socket &operator<<(const socket &, const std::pair<first_type, second_type> &);

	template<class first_type, class second_type>
		const socket &operator>>(const socket &, std::pair<first_type, second_type> &);

	const socket &operator<<(const socket &, const std::vector<unsigned char> &);
	const socket &operator<<(const socket &, command);

	bool tryrecv(const socket &, unsigned &);

	void recvall(const socket &, std::vector<unsigned char> &);
	void sendall(const socket &, const std::vector<unsigned char> &);
}

class p2pvdn::socket {
	int socketfd;

	public:
	socket(int socketfd=-1)
		: socketfd(socketfd)
	{
	}

	socket(const socket &socket) throw(network_error)
		: socketfd(-1)
	{
		if (socket.socketfd != -1) {
			socketfd = dup(socket.socketfd);
			if (socketfd == -1)
				switch (errno) {
					case EBADF:
						throw invalid_socket();
					case EMFILE:
						throw file_descriptor_limit_reached();
					default:
						throw network_error();
				}
		}
	}

	socket(socket &&socket) throw()
		: socketfd(socket.socketfd)
	{
		socket.socketfd = -1;
	}

	socket &operator=(const socket &socket)
	{
		*this = { socket };
		return *this;
	}

	socket &operator=(socket &&socket) throw(network_error)
	{
		this->close();

		socketfd = socket.socketfd;
		socket.socketfd = -1;

		return *this;
	}

	~socket() throw()
	{
		if (socketfd != -1)
			try {
				this->close();
			}
			catch (network_error) {
				std::clog << "error closing socket\n";
			}
	}

	operator command() const
	{
		command buffer;
		recvall(&buffer, sizeof(buffer));

		return buffer;
	}

	operator std::vector<unsigned char>() const
	{
		unsigned size;
		*this >> size;
		std::vector<unsigned char> vector(size);
		this->recvall(&vector[0], size);

		return vector;
	}

	int value() const
	{
		return socketfd;
	}

	void recvall(void *data, size_t size) const;
	bool tryrecv(void *data, size_t size) const;
	unsigned recv(void *data, size_t size) const;
	void sendall(const void *data, size_t size) const;
	void close();
};

inline const p2pvdn::socket &p2pvdn::operator<<(const socket &socket, unsigned src)
{
	uint32_t buf = htonl(src);
	socket.sendall(&buf, sizeof(buf));

	return socket;
}

inline const p2pvdn::socket &p2pvdn::operator>>(const socket &socket, unsigned &dest)
{
	uint32_t buf;
	socket.recvall(&buf, sizeof(buf));
	dest = ntohl(buf);

	return socket;
}

inline const p2pvdn::socket &p2pvdn::operator<<(const socket &socket, unsigned char src)
{
	socket.sendall(&src, sizeof(src));

	return socket;
}

inline const p2pvdn::socket &p2pvdn::operator>>(const socket &socket, unsigned char &dest)
{
	socket.recvall(&dest, sizeof(dest));

	return socket;
}

template<class type>
inline const p2pvdn::socket
&p2pvdn::operator<<(const socket &socket, const std::vector<type> &vector)
{
	socket << static_cast<unsigned>(vector.size());
	for (auto iterator = vector.begin(); iterator != vector.end(); iterator++)
		socket << *iterator;

	return socket;
}

inline const p2pvdn::socket &p2pvdn::operator<<(const socket &socket,
		const std::vector<unsigned char> &vector)
{
	unsigned size = vector.size();
	socket << size;
	socket.sendall(&vector[0], size); 

	return socket;
}

template<class type>
inline const p2pvdn::socket &p2pvdn::operator>>(const socket &socket, std::vector<type> &vector)
{
	unsigned size;
	socket >> size;
	vector = std::vector<type>(size);
	for (auto iterator = vector.begin(); iterator != vector.end(); iterator++)
		socket >> *iterator;

	return socket;
}

template<class first_type, class second_type>
inline const p2pvdn::socket &p2pvdn::operator<<(const socket &socket,
	const std::pair<first_type, second_type> &pair)
{
	socket << pair.first << pair.second;

	return socket;
}

template<class first_type, class second_type>
inline const p2pvdn::socket &p2pvdn::operator>>(const socket &socket,
	std::pair<first_type, second_type> &pair)
{
	socket >> pair.first >> pair.second;

	return socket;
}

inline const p2pvdn::socket &p2pvdn::operator<<(const socket &socket, command command)
{
	socket.sendall(&command, sizeof(command));

	return socket;
}

inline bool p2pvdn::tryrecv(const socket &socket, unsigned &dest)
{
	uint32_t buf;
	if (!socket.tryrecv(&buf, sizeof(buf)))
		return false;

	dest = ntohl(buf);
	return true;
}

inline void p2pvdn::recvall(const socket &socket, std::vector<unsigned char> &vector)
{
	socket.recvall(&vector[0], vector.size());
}

inline void p2pvdn::sendall(const socket &socket, const std::vector<unsigned char> &vector)
{
	socket.sendall(&vector[0], vector.size());
}

#endif
