#! /bin/bash -e

VDND="../vdnd/vdnd"
DISPERSE="../disperse/disperse"
GET_FILE="../get_file/get_file"

usage() {
	echo "test_replenishment.sh num_nodes"
}

error() {
	echo error: "$1" >&2
	exit 1
}

node_address() {
	declare -i node_num="$1"+1 octet1 octet2 octet3

	((octet1=$node_num%256))
	((node_num/=256))
	((octet2=$node_num%256))
	((node_num/=256))
	((octet3=$node_num%256))
	echo "127.$octet3.$octet2.$octet1"
}

create_network() {
	declare -i node_num="$1"
	declare -a vdnd_args=( -a $(node_address $node_num) ) 

	if [ -n "$port" ]; then
		vdnd_args=( "${vdnd_args[@]}" -p $port )
	fi
	eval "$VDND" "${vdnd_args[@]}" \> /dev/null \&
	echo $!
}

join_network() {
	declare -i node_num="$1" join_node_num="$2"
	declare -a vdnd_args=( -a $(node_address $node_num)
			-j $(node_address ${join_node_num}) )

	if [ -n "$port" ]; then
		vdnd_args=( "${vdnd_args[@]}" -p $port -P $port )
	fi
	eval "$VDND" "${vdnd_args[@]}" \> /dev/null \&
	echo $!
}

init_network() {
	declare -i num_nodes="$1" node_num

	num_nodes="$1"
	for ((node_num=0; node_num<num_nodes; node_num++)); do
		vdnd_args=( -a $(node_address $node_num ) )
		if [ $node_num -eq 0 ]; then
			create_network $node_num
		else
			join_network $node_num $(($node_num-1))
		fi
		sleep 1
	done
}

kill_node() {
	declare -i pid="$1"

	kill $pid
}

disperse_file() {
	declare file="$1"
	declare -i node_num="$2"
	declare -a disperse_args=( "$file" -a $(node_address $node_num) )

	if [ -n "$port" ]; then
		disperse_args=( "${disperse_args[@]}" -p $port )
	fi

	"$DISPERSE" "${disperse_args[@]}"
}

get_file() {
	declare file="$1"
	declare -i node_num="$2"
	declare -a get_file_args=( "$file" -a $(node_address $node_num) )

	if [ -n "$port" ]; then
		get_file_args=( "${get_file_args[@]}" -p $port )
	fi

	timeout 10s $GET_FILE "${get_file_args[@]}"
}

set_traps() {
	local signal="$1"
	shift
	declare -a traps=( "$@" )
	local IFS=";"

	trap "${traps[*]}" "$signal"
}

args="$(getopt -o p: --long port: -- "$@")"
eval set -- "$args"
unset args
while true; do
	case "$1" in
		-p|--port)
			declare -i port="$2"
			shift 2
			;;
		--)
			shift
			break
			;;
		*)
			error "invalid option"
			;;
	esac
done
port=${port:-""}

if [ $# -ne 1 ]; then
	usage
	error "incorrect number of arguments"
fi

declare -i num_nodes="$1"
if [ $num_nodes -lt 2 ]; then
	error "must use at least two nodes"
fi
declare -a network=( $(seq 0 $num_nodes) ) pids

if [ ! -x "$VDND" ]; then
	error "vdnd not found"
fi
if [ ! -x "$DISPERSE" ]; then
	error "disperse not found"
fi

declare -a exit_traps

pids=( $(init_network $num_nodes) )

declare file hash
file="$(mktemp)"
exit_traps=( "${exit_traps[@]}" "rm -f $file" )
set_traps EXIT "${exit_traps[@]}"
dd if=/dev/urandom of="$file" count=$((2**10)) > /dev/null
hash="$(sha1sum "$file")"
disperse_file "$file" 0
echo "$hash"

declare -i random_node join_node node_num
for ((node_num=$num_nodes; node_num++; )); do
	random_node=$(($RANDOM % $num_nodes))
	if [ $random_node -ne 0 ]; then
		join_node=$random_node-1
	else
		join_node=$random_node+1
	fi
	kill_node ${pids[$random_node]}
	sleep 1
	network[$random_node]=$node_num
	pids[$random_node]=$(join_network $node_num ${network[$join_node]})
	sleep 1
	if ! get_file "$file" ${network[$random_node]}; then
		echo "failed to get file"
		break
	fi
	sha1sum "$file"
done
echo "$(($node_num-$num_nodes)) iterations"
for pid in "${pids[@]}"; do
	kill $pid
done
exit 0
