#include "vdnd.h"
#include "server.h"

#include <chord.h>

#include <exception>
#include <thread>

#include <cstdlib>
#include <csignal>

#include <getopt.h>

namespace vdnd {
	struct opts {
		int join;
		std::string local_node;
		std::string local_service;
		std::string node;
		std::string service;
		int hash_size;
		int sleep_time;

		opts()
			: join(0), local_node("127.0.0.1"), local_service("50000"),
			service("50000"), hash_size(20), sleep_time(10)
		{
		}
	};

	static void parse_options(int argc, char *argv[], struct opts &vdn_opts)
	{
		const struct option opts[] = {
			{ "address", required_argument, NULL, 'a' },
			{ "port", required_argument, NULL, 'p' },
			{ "join", required_argument, NULL, 'j' },
			{ "hash_size", required_argument, NULL, 's' },
			{ "remote_port", required_argument, NULL, 'P' },
			{ "sleep_time", required_argument, NULL, 't' },
			{ 0, 0, 0, 0 }
		};
		const char short_opts[] = "a:p:j:s:P:t:";

		int opt;
		while ((opt = getopt_long(argc, argv, short_opts, opts, NULL)) != -1) {
			switch (opt) {
				case 'a':
					vdn_opts.local_node = optarg;
					break;
				case 'p':
					vdn_opts.local_service = optarg;
					break;
				case 'j':
					vdn_opts.join = 1;
					vdn_opts.node = optarg;
					break;
				case 'P':
					vdn_opts.service = optarg;
					break;
				case 's':
					vdn_opts.hash_size = atoi(optarg);
					break;
				case 't':
					vdn_opts.sleep_time = atoi(optarg);
					break;
				default:
					throw std::invalid_argument("unrecognized option");
			}
		}
	}

	static void handle_signals(void)
	{
		int retval = 0;
		int exit, signal;
		sigset_t handled_signals;

		retval = sigemptyset(&handled_signals);
		if (!retval)
			retval = sigaddset(&handled_signals, SIGINT);

		if (!retval)
			retval = sigaddset(&handled_signals, SIGTERM);

		for (exit = 0; !retval && !exit; ) {
			retval = sigwait(&handled_signals, &signal);
			if (!retval)
				switch (signal) {
					case SIGINT:
					case SIGTERM:
						exit = 1;
						break;
					default:
						retval = -1;
				}
		}

		if (retval)
			throw std::exception();
	}

	static void block_signals(void)
	{
		int retval = 0;
		sigset_t blocked_signals;

		retval = sigemptyset(&blocked_signals);
		if (!retval)
			retval = sigaddset(&blocked_signals, SIGINT);

		if (!retval)
			retval = sigaddset(&blocked_signals, SIGTERM);

		if (!retval)
			retval = pthread_sigmask(SIG_BLOCK, &blocked_signals, NULL);

		if (retval)
			throw std::exception();
	}
}

vdnd::ctx::ctx(const opts &opts)
	: sleep_time(opts.sleep_time)
{
}

int main(int argc, char *argv[])
{
	int retval = 0;
	struct chord_network *chord_network;
	struct chord_opts chord_opts;

	vdnd::block_signals();

	vdnd::opts vdn_opts;
	vdnd::parse_options(argc, argv, vdn_opts);

	vdnd::ctx ctx(vdn_opts);
	chord_opts.join = vdn_opts.join ? 1: 0;
	chord_opts.local_node = vdn_opts.local_node.c_str();
	chord_opts.local_service = vdn_opts.local_service.c_str();
	chord_opts.node = vdn_opts.node.c_str();
	chord_opts.service = vdn_opts.service.c_str();
	chord_opts.hash_size = vdn_opts.hash_size;
	chord_opts.num_successors = 0;
	chord_opts.handler = vdnd::server;
	chord_opts.arg = &ctx;
	retval = chord_start(&chord_opts, &chord_network);
	if (retval)
		return EXIT_FAILURE;

	vdnd::handle_signals();
	chord_stop(chord_network);

	return EXIT_SUCCESS;
}
